/****************************************************************************
 **
 ** Copyright (C) 2010 Arzel Jrme <myst6re@gmail.com>
 **
 ** This program is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 **
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with this program.  If not, see <http://www.gnu.org/licenses/>.
 ****************************************************************************/

#include "SaveData.h"

SaveData::SaveData(int id, QByteArray data, QByteArray MC_header)
	: _id(id), _isFF8(false), _isDelete(false), _isInvalid(false)
{
	_isDelete = data.isEmpty();// 0xa1-0xa2-0xa3 : deleted | 0xa0 : formated
	bool correctSize = data.size() >= FF8SAVE_SIZE;

	setMC_header(MC_header);

	if(correctSize && charger(data))
	{
		_isFF8 = true;
	}
	else if(!_isDelete)
	{
//		qDebug() << data.size() << data.startsWith("SC");
		if(correctSize && data.startsWith("SC"))
		{
			_header = data.left(96);
			switch((quint8)data.at(2)) {
			case 0x11:
				_icon.setAll(data.mid(96,160));
				break;
			case 0x12:
				_icon.setAll(data.mid(96,288), 2);
				break;
			case 0x13:
				_icon.setAll(data.mid(96,416), 3);
				break;
//			default:
//				qDebug() << (quint8)data.at(2);
			}
		}
		else
			_isInvalid = true;
	}
}

SaveData::~SaveData()
{
}

QByteArray SaveData::MC_header()
{
	return _MC_header;
}

QByteArray SaveData::saveMC_header()
{
	if(hasMC_header()) {
		_MC_header = _MC_header.leftJustified(127, '\x00', true);
		return _MC_header + (char)xorByte(_MC_header.constData());
	}
	else {
		return QByteArray("\xa0\x00\x00\x00\x00\x00\x00\x00\xff\xff", 10).append(QByteArray(118, '\x00'));
	}
}

void SaveData::setMC_header(QByteArray MC_header)
{
	_MC_header = MC_header;

	if(hasMC_header()) {
		_isDelete = (quint8)MC_header.at(0) >> 4 == 0xa;
		_freq_value = MC_header.at(11)=='E' ? 50 : 60;
	}
	else {
		_isDelete = false;
		_freq_value = 60;
	}
}

void SaveData::setMC_header_id(int id)
{
	if(hasMC_header()) {
		_MC_header.replace(26, 2, QString("%1").arg(id, 2, 10, QChar('0')).toLatin1());
	}
}

QByteArray SaveData::header()
{
	return _header;
}

QPixmap SaveData::icon(bool chocobo_world_icon)
{
	return _icon.icon(chocobo_world_icon);
}

SaveIcon *SaveData::saveIcon()
{
	return &_icon;
}

HEADER SaveData::descData()
{
	return _descData;
}

void SaveData::setDescData(HEADER descData)
{
	_descData = descData;
}

MAIN SaveData::mainData()
{
	return _mainData;
}

void SaveData::setMainData(MAIN data)
{
	_mainData = data;
}

int SaveData::freq_value()
{
	return _freq_value;
}

int SaveData::id()
{
	return _id;
}

bool SaveData::isFF8()
{
	return _isFF8;
}

bool SaveData::isDelete()
{
	return _isDelete;
}

bool SaveData::isInvalid()
{
	return _isInvalid;
}

bool SaveData::hasMC_header()
{
	return !_MC_header.isEmpty();
}

bool SaveData::exportPC(QString path)
{
	QFile fic(path);
	if(!fic.open(QIODevice::WriteOnly))	return false;
	
	QByteArray data = LZS::compress(sauver());
	quint32 size = data.size();
	
	fic.write((char *)&size, 4);
	fic.write(data);
	fic.close();

	return true;
}

void SaveData::restore()
{
	_isDelete = false;
	_MC_header.replace(0, 10, QByteArray("\x51\x00\x00\x00\x00\x20\x00\x00\xff\xff", 10));
	_MC_header.replace(28, 99, QByteArray(99, '\x00'));
}

QString SaveData::getShortDescription()
{
	QTextCodec *codec = QTextCodec::codecForName(QByteArray("Shift-JIS"));
	if(!_header.isEmpty() && codec!=0) {
		QByteArray desc_array = _header.mid(4,64);
		int index;
		if((index = desc_array.indexOf('\x00')) != -1) {
			desc_array.truncate(index);
		}
		return codec->toUnicode(desc_array);
	}
	return QString();
}

bool SaveData::charger(QByteArray data)
{
	const char *access_data = data.constData();
	//ff8
	quint16 ff8;
	memcpy(&ff8, &access_data[386], 2);
	if(ff8 != 0x8FF)		return false;
	
	//Avant le checksum
	_header = data.left(96);
	_icon.setAll(data.mid(96,288));

	if(sizeof(_descData)!=76) {
		qWarning() << "error desc structure" << sizeof(_descData);
		return false;
	}

	memcpy(&_descData, &access_data[388], sizeof(_descData));

	if(sizeof(_mainData)!=4944) {
		qWarning() << "error main structure" << sizeof(_mainData);
		return false;
	}

	memcpy(&_mainData, &access_data[464], sizeof(_mainData));

	//qDebug() << data.mid(457,7).toHex();

	return true;
}

QByteArray SaveData::sauver()
{
	quint8 leader;
	quint8 perso1 = _mainData.misc.party[0];
	quint8 perso2 = _mainData.misc.party[1];
	quint8 perso3 = _mainData.misc.party[2];
	
	if(perso1 != 0xFF && perso1 < 8)
		leader = perso1;
	else if(perso2 != 0xFF && perso2 < 8)
		leader = perso2;
	else if(perso3 != 0xFF && perso3 < 8)
		leader = perso3;
	else
		leader = 0;
	if(perso1 != 0xFF && perso1 < 8)	perso1 = _mainData.persos[perso1].ID;
	if(perso2 != 0xFF && perso2 < 8)	perso2 = _mainData.persos[perso2].ID;
	if(perso3 != 0xFF && perso3 < 8)	perso3 = _mainData.persos[perso3].ID;

	_descData.party[0] = perso1;
	_descData.party[1] = perso2;
	_descData.party[2] = perso3;
	
	_descData.nivLeader = _mainData.persos[leader].exp/1000 + 1;
	_descData.hpLeader = _mainData.persos[leader].current_HPs;
	_descData.gils = _mainData.misc.argent;
	_descData.time = _mainData.stats.game_time;

	quint16 checksum = calcChecksum((char *)&_mainData);//On calcule le checksum  partir de la partie gf

	QByteArray ret;
	ret.append(_header.left(4));//SC+icon frames+slot count
	ret.append("\x82\x65\x82\x65\x82\x57\x81\x6D", 8);// FF8[
	ret.append(FF8Text::numToBiosText(_id+1, 2));// II
	ret.append("\x81\x6E\x81\x5E",4);// ]/
	ret.append(FF8Text::numToBiosText(Config::hour(_mainData.stats.game_time, _freq_value), 2));// HH
	ret.append("\x81\x46",2);// :
	ret.append(FF8Text::numToBiosText(Config::min(_mainData.stats.game_time, _freq_value), 2));// MM
	ret.append(_header.right(66));
	ret.append(_icon.sauver());
	ret.append((char *)&checksum, 2);
	ret.append("\xFF\x08", 2);
	ret.append((char *)&_descData, sizeof(_descData));
	ret.append((char *)&_mainData, sizeof(_mainData));
	ret.append((char *)&checksum, 2);
	ret.append(QByteArray(2782,'\x00'));

	if(ret.size()!=SAVE_SIZE)	qWarning() << "Error saved save size" << ret.size() << SAVE_SIZE;
	
	return ret;
}

bool SaveData::isJp()
{
	return _MC_header.isEmpty() ? false : _MC_header.at(11)=='I';
}

QString SaveData::perso(quint8 index)
{
	switch(index) {
	case 0:				return FF8Text::toString(_descData.squall, isJp());
	case 4:				return FF8Text::toString(_descData.linoa, isJp());
	case 12:			return FF8Text::toString(_mainData.misc.cronos, isJp());
	case 14:			return FF8Text::toString(_descData.boko, isJp());
	case 15:			return FF8Text::toString(_descData.angelo, isJp());
	default:
		if(index < 16)	return Data::names.at(index);
		return QString();
	}
}

quint8 SaveData::xorByte(const char *data)
{
	register quint8 xorByte=0;
	for(int i=0 ; i<127 ; ++i)
		xorByte ^= data[i];
	return xorByte;
}

quint16 SaveData::calcChecksum(const char *data)
{
	register quint16 crc = 0xFFFF, cur;

	for(cur=0 ; cur<4944 ; ++cur)
		crc = crcTab[(crc >> 8) ^ (quint8)*data++] ^ (crc << 8);

	return ~crc;
}

const quint16 SaveData::crcTab[256] = {0,4129,8258,12387,16516,20645,24774,28903,33032,37161,41290,45419,49548,53677,57806,61935,4657,528,12915,8786,21173,17044,29431,25302,37689,33560,45947,41818,54205,50076,62463,58334,9314,13379,1056,5121,25830,29895,17572,21637,42346,46411,34088,38153,58862,62927,50604,54669,13907,9842,5649,1584,30423,26358,22165,18100,46939,42874,38681,34616,63455,59390,55197,51132,18628,22757,26758,30887,2112,6241,10242,14371,51660,55789,59790,63919,35144,39273,43274,47403,23285,19156,31415,27286,6769,2640,14899,10770,56317,52188,64447,60318,39801,35672,47931,43802,27814,31879,19684,23749,11298,15363,3168,7233,60846,64911,52716,56781,44330,48395,36200,40265,32407,28342,24277,20212,15891,11826,7761,3696,65439,61374,57309,53244,48923,44858,40793,36728,37256,33193,45514,41451,53516,49453,61774,57711,4224,161,12482,8419,20484,16421,28742,24679,33721,37784,41979,46042,49981,54044,58239,62302,689,4752,8947,13010,16949,21012,25207,29270,46570,42443,38312,34185,62830,58703,54572,50445,13538,9411,5280,1153,29798,25671,21540,17413,42971,47098,34713,38840,59231,63358,50973,55100,9939,14066,1681,5808,26199,30326,17941,22068,55628,51565,63758,59695,39368,35305,47498,43435,22596,18533,30726,26663,6336,2273,14466,10403,52093,56156,60223,64286,35833,39896,43963,48026,19061,23124,27191,31254,2801,6864,10931,14994,64814,60687,56684,52557,48554,44427,40424,36297,31782,27655,23652,19525,15522,11395,7392,3265,61215,65342,53085,57212,44955,49082,36825,40952,28183,32310,20053,24180,11923,16050,3793};
